if (!ko) var ko = {};
if (!ko.collaboration) ko.collaboration = {};

xtk.include("treeview");

let SessionsTreeView = function() {
    this._rows = [];
}

SessionsTreeView.prototype = new xtk.dataTreeView();
SessionsTreeView.prototype.constructor = SessionsTreeView;

SessionsTreeView.prototype.getSelectedIndices = function() {
    // Note: seltype is currently "single". A lot of the code below in
    // ko.collaboration.sessions relies on that.
    var indices = [];
    var len = this.selection.getRangeCount()
    for (let i = 0; i < len; i++) {
        let start = {};
        let end = {};
        this.selection.getRangeAt(i, start, end);
        for (let j=start.value; j <= end.value; j++)
            indices.push(j);
    }
    return indices;
};

SessionsTreeView.prototype.getCellText = function(row, column) {
    var col_index = column.index;
    return this._rows[row].getCellText(col_index);
};

SessionsTreeView.prototype.isSeparator = function(row) {
    return this._rows[row].isSeparator();
};

SessionsTreeView.prototype.isContainer = function(row) {
    return this._rows[row].isContainer();
};

SessionsTreeView.prototype.isContainerOpen = function(row) {
    return this._rows[row].isContainerOpen();
};

SessionsTreeView.prototype.isContainerEmpty = function(row) {
    return this._rows[row].isContainerEmpty();
};

SessionsTreeView.prototype.getParentIndex = function(row) {
    var rowLevel = this.getLevel(row);
    if (rowLevel == 0)
        return -1;
    for (let i = row; i >= 0; i--)
        if (this.getLevel(i) < rowLevel)
            return i;
    return -1;
};

SessionsTreeView.prototype.getLevel = function(row) {
    return this._rows[row].getLevel();
};

SessionsTreeView.prototype.hasNextSibling = function(row, after) {
    var thisLevel = this.getLevel(row);
    var l = this._rows.length;
    for (var i = after + 1; i < l; i++) {
        var nextLevel = this.getLevel(i);
        if (nextLevel == thisLevel) return true;
        if (nextLevel < thisLevel) break;
    }
    return false;
};

SessionsTreeView.prototype.toggleOpenState = function(row) {
    var node = this._rows[row];
    if (!node.isContainer())
        return;
    var oldRows = node.flatRows();
    if (node.isContainerOpen()) {
        node.isOpen = false;
    }
    else {
        node.isOpen = true;
    }
    var newRows = node.flatRows();
    var firstSlice = this._rows.slice(0, row);
    var lastSlice = this._rows.slice(row + oldRows.length);
    this._rows = firstSlice.concat(newRows).concat(lastSlice);
    this.tree.rowCountChanged(row+1, newRows.length - oldRows.length);
};

SessionsTreeView.prototype.getImageSrc = function(row, column) {
    var node = this._rows[row];
    return node.getImageSrc(column);
}

SessionsTreeView.prototype.getCellProperties = function(row, column, properties) {
    var node = this._rows[row];
    node.getCellProperties(column, properties);
}

SessionsTreeView.prototype.getIndex = function(node) {
    for (let index = 0; index < this._rows.length; index++) {
        if (this._rows[index] == node) return index;
    }
    return -1;
}

SessionsTreeView.prototype.getRootLevelNode = function(index) {
    while(this.getLevel(index) > 0) {
        index = this.getParentIndex(index);
    }
    return this._rows[index];
}

SessionsTreeView.prototype.deleteRow = function(row) {
    return this._rows[row].deleteNode();
}


let TreeNode = function(cells, level, isOpen, parentNode) {
    this.cells = cells;
    this.level = level;
    this.isOpen = isOpen;
    this.type = "node";
    this.parent = parentNode;
};

TreeNode.prototype.isSeparator = function() {
    return false;
}

TreeNode.prototype.getCellText = function(col_idx) {
    return String(this.cells[col_idx]);
};

TreeNode.prototype.getChildren = function() {
    return null;
}

TreeNode.prototype.isContainer = function() {
    return !!this.getChildren();
};

TreeNode.prototype.isContainerOpen = function() {
    return this.isOpen;
};

TreeNode.prototype.isContainerEmpty = function() {
    var children = this.getChildren();
    for (var key in children) {
        if (children.hasOwnProperty(key)) return false;
    }
    return true;
};

TreeNode.prototype.getLevel = function() {
    return this.level;
};

TreeNode.prototype.getImageSrc = function(row, column) {};

TreeNode.prototype.getCellProperties = function(column, properties) {};

TreeNode.prototype.flatRows = function() {
    var rows = [];
    rows.push(this);
    if (this.isContainer() && this.isContainerOpen() && !this.isContainerEmpty()) {
        for (let [foo, childNode] in Iterator(this.getChildren())) {
            rows = rows.concat(childNode.flatRows());
        }
    }
    return rows;
}

TreeNode.prototype.updateNodeData = function(data) {
    throw "Child class must implement updateNodeData";
}

TreeNode.prototype.invalidate = function() {
    let myIndex = this.treeView.getIndex(this);
    this.tree.invalidateRow(myIndex);
}

TreeNode.prototype.hasOwnerPrivileges = function() {
    // Returns true if the current user has Owner privileges on this resource
    return false;
}

TreeNode.prototype.canBeRenamed = function() {
    // Returns true if 'Rename' in context menu should be enabled.
    return false;
}

TreeNode.prototype.canBeDeleted = function() {
    // Returns true if 'Delete' in context menu should be enabled.
    return false;
}

TreeNode.prototype.deleteNode = function() {
    return false;
}


let SeparatorNode = function(level, parentNode) {
    TreeNode.call(this, [""], level, false, parentNode);
}

SeparatorNode.prototype.__proto__ = TreeNode.prototype;

SeparatorNode.prototype.isSeparator = function() {
    return true;
}

// Privilege levels on sessions. Currently only 2 and 3 are in use.
let SESSION_PRIVILEGES = {
    0: 'NONE',
    1: 'READ',
    2: 'WRITE',
    3: 'OWNER'
};

let SessionNode = function(id, name) {
    TreeNode.call(this, [], 0, false, null);
    this.type = "session";
    this.id = id;
    this.name = name;
    this.separatorNode = new SeparatorNode(1, this);
    this.documentNodes = {};
    this.collaboratorNodes = {};
    this.userPrivilege = 'READ';
};

SessionNode.prototype.__proto__ = TreeNode.prototype;

SessionNode.prototype.getCellText = function(col_idx) {
    if (col_idx != 0) return null;
    return this.name;
};

SessionNode.prototype.getChildren = function() {
    var children = [];
    for each (let node in this.documentNodes) {
        children.push(node);
    }
    if (children.length > 0) {
        children.push(this.separatorNode);
    }
    for each (let node in this.collaboratorNodes) {
        children.push(node);
    }
    return children;
}

SessionNode.prototype.getCellProperties = function(column, properties) {
    var aserv=Components.classes["@mozilla.org/atom-service;1"].
              getService(Components.interfaces.nsIAtomService);
    properties.AppendElement(aserv.getAtom("sessionNode"));
};

/**
 * Takes session data returned by the API and updates its children with it.
 * Returns true if something's changed and the tree needs to be updated and
 * false otherwise.
 */
SessionNode.prototype.updateNodeData = function(sessionData) {
    var changed = false;
    if (typeof sessionData == 'object') {
        if (sessionData.hasOwnProperty('name')) {
            if (this.name !== sessionData.name) {
                this.name = sessionData.name;
                changed = true;
            }
        }
        if (sessionData.hasOwnProperty('texts')) {
            changed |= this.updateDocuments(sessionData.texts);
        }
        if (sessionData.hasOwnProperty('privileges')) {
            changed |= this.updateCollaborators(sessionData.privileges);
        }
    }
    return changed;
};

SessionNode.prototype.updateDocuments = function(documents) {
    var changed = false;
    // Remove deleted nodes
    for each (let [id, node] in Iterator(this.documentNodes)) {
        if (!(id in documents)) {
            delete this.documentNodes[id];
            changed = true;
        }
    }
    // update existing and add new nodes
    for each (let [id, title] in Iterator(documents)) {
        if (id in this.documentNodes) {
            changed |= this.documentNodes[id].updateNodeData(title)
        } else {
            this.documentNodes[id] = new DocumentNode(id, title, this);
            changed = true;
        }
    }
    return changed;
};

SessionNode.prototype.updateCollaborators = function(collaborators) {
    var changed = false;
    var myUserId = ko.collaboration.service.ssoUserId || '';
    // Remove deleted nodes
    for each (let [userId, node] in Iterator(this.collaboratorNodes)) {
        if (!(userId in collaborators)) {
            delete this.collaboratorNodes[userId];
            changed = true;
        }
    }
    // update existing and add new nodes
    for each (let [userId, userData] in Iterator(collaborators)) {
        // If this is the local user, update userPrivilege of the SessionNode
        if (userId == myUserId) {
            var myPrivilege = userData[1];
            myPrivilege = SESSION_PRIVILEGES[myPrivilege];
            if (myPrivilege)
                this.userPrivilege = myPrivilege;
        }
        // Update the collaborator Node
        if (userId in this.collaboratorNodes) {
            // A single collaborator node shouldn't change.
        } else {
            let userName = userData[0];
            let privilege = userData[1];
            this.collaboratorNodes[userId] = new CollaboratorNode(userId, userName, privilege, this);
            changed = true;
        }
    }
    return changed;
};

SessionNode.prototype.hasOwnerPrivileges = function() {
    return this.userPrivilege == 'OWNER';
};

SessionNode.prototype.canBeRenamed = function() {
    return this.hasOwnerPrivileges();
};

SessionNode.prototype.canBeDeleted = function() {
    return this.hasOwnerPrivileges();
};

SessionNode.prototype.deleteNode = function() {
    ko.collaboration.api.deleteSessionAsync(this.id, function() {
        ko.collaboration.sessions.refresh();
    }, null);
    return true;
};


let DocumentNode = function(id, name, session) {
    TreeNode.call(this, null, 2, false, session);
    this.type = "document";
    this.id = id;
    this.name = name;
};

DocumentNode.prototype.__proto__ = TreeNode.prototype;

DocumentNode.prototype.getCellText = function(colIdx) {
    if (colIdx != 0)
        return undefined;
    return this.name;
}

DocumentNode.prototype.getImageSrc = function(row, column) {
    var filename = this.name;
    // At least on OS X filenames without an extension don't get an icon.
    // Suffix those with a fake file extension.
    var extRe = /^.*\..+$/;
    if (!extRe.test(filename)) {
        filename += ".txt";
    }
    return "moz-icon://" + filename + "?size=16";
}

DocumentNode.prototype.getCellProperties = function(column, properties) {
    var aserv=Components.classes["@mozilla.org/atom-service;1"].
              getService(Components.interfaces.nsIAtomService);
    properties.AppendElement(aserv.getAtom("documentNode"));
};

DocumentNode.prototype.hasOwnerPrivileges = function() {
    // Returns true if the current user has Owner privileges on this resource
    return this.parent.hasOwnerPrivileges();
}

DocumentNode.prototype.canBeRenamed = function() {
    return true;
};

DocumentNode.prototype.canBeDeleted = function() {
    return true;
};

DocumentNode.prototype.deleteNode = function() {
    ko.collaboration.api.deleteTextAsync(this.id, function() {
        ko.collaboration.sessions.refresh();
    }, null);
    return true;
};

DocumentNode.prototype.updateNodeData = function(name) {
    if (this.name != name) {
        this.name = name;
        return true;
    }
    return false;
}


let CollaboratorNode = function(id, name, privilege, session) {
    let cellText = name;
    if (privilege == 3) cellText += " (Owner)";
    TreeNode.call(this, [cellText], 2, false, session);
    this.type = "collaborator";
    this.id = id;
    this.name = name;
    this.privilege = privilege;
};

CollaboratorNode.prototype.__proto__ = TreeNode.prototype;

CollaboratorNode.prototype.getCellProperties = function(column, properties) {
    var aserv=Components.classes["@mozilla.org/atom-service;1"].
              getService(Components.interfaces.nsIAtomService);
    properties.AppendElement(aserv.getAtom("collaboratorNode"));
};

CollaboratorNode.prototype.hasOwnerPrivileges = function() {
    return this.parent.hasOwnerPrivileges();
};

CollaboratorNode.prototype.canBeRenamed = function() {
    return false;
};

CollaboratorNode.prototype.canBeDeleted = function() {
    return this.hasOwnerPrivileges();
};

CollaboratorNode.prototype.deleteNode = function() {
    let sessionNode = this.parent;
    if (sessionNode instanceof SessionNode) {
        ko.collaboration.api.revokeAccessAsync(this.id, sessionNode.id, function() {
            ko.collaboration.sessions.refresh();
        }, null);
        return true;
    }
    return false;
};


/**
 * Collaboration Session Tree
 */
ko.collaboration.sessions = {
    init: function() {
        this.tree = document.getElementById('collab-sessions-tree');
        this.treeView = new SessionsTreeView();
        this.tree.treeBoxObject.view = this.treeView;
        this.sessions = {};
    },

    /**
     * Takes a hash of session ids and session data, usually the returned data
     * of a server request, compares it to the stored session data, and
     * updates the tree accordingly.
     */
    _updateSessions: function(newSessions) {
        for each (let [sessionId, sessionData] in Iterator(newSessions)) {
            if (sessionId in this.sessions) {
                // existing session node, update
                this._updateSession(sessionId, sessionData);
            } else {
                // New session, that has not been created by this user, but
                // this user has been added _to_ that session by someone else
                this.addSession(sessionId, sessionData, false);
            }
        }

        for each (let [sessionId, sessionNode] in Iterator(this.sessions)) {
            if (!(sessionId in newSessions)) {
                this.removeSession(sessionId);
            }
        }
    },

    _updateSession: function(sessionId, sessionData) {
        if (sessionId in this.sessions) {
            var sessionNode = this.sessions[sessionId];
            var oldRows = sessionNode.flatRows();
            var nodeChanged = sessionNode.updateNodeData(sessionData);
            if (nodeChanged) {
                var newRows = sessionNode.flatRows();
                var rowIndex = this.treeView.getIndex(sessionNode);

                var firstSlice = this.treeView._rows.slice(0, rowIndex);
                var lastSlice = this.treeView._rows.slice(rowIndex + oldRows.length);
                this.treeView._rows = firstSlice.concat(newRows).concat(lastSlice);
                // FIXME This is not really how you should use rowCountChanged,
                // so beware of strange side-effects w/ multiple selections
                this.treeView.tree.rowCountChanged(rowIndex + 1,
                                            newRows.length - oldRows.length);
                this.treeView.tree.invalidateRange(rowIndex,
                                                   rowIndex + newRows.length)
            }
        }
    },

    addSession: function(sessionId, sessionData, nodeOpen) {
        if (sessionId in this.sessions) {
            // Don't add a session twice (e.g. when refresh() is called by an
            // aysnchronous callback before addSession() is invoked)
            this.removeSession(sessionId);
        }
        var sessionNode = new SessionNode(sessionId, sessionData["name"]);
        sessionNode.isOpen = !!nodeOpen;
        sessionNode.updateNodeData(sessionData);
        var nodeRows = sessionNode.flatRows();
        var index = this.treeView._rows.length;
        this.treeView._rows = this.treeView._rows.concat(nodeRows);
        this.treeView.tree.rowCountChanged(index, nodeRows.length);
        this.sessions[sessionId] = sessionNode;
    },

    removeSession: function(sessionId) {
        if (sessionId in this.sessions) {
            let sessionNode = this.sessions[sessionId];
            let index = this.treeView.getIndex(sessionNode);
            if (index < 0) {
                // Reset the whole tree
                this.refresh();
                return;
            }
            let rows = sessionNode.flatRows();
            this.treeView._rows.splice(index, rows.length);
            this.treeView.tree.rowCountChanged(index, -rows.length);
            delete this.sessions[sessionId];
        }
    },

    refresh: function() {
        let successCallback =
            ko.collaboration.utils.bind(this._updateSessions, this);
        let errorCallback = function() {};
        ko.collaboration.api.getSessionsAsync(successCallback, errorCallback)
    },

    createSession: function() {
        var sessionName = ko.dialogs.prompt("Enter a session name", null, null,
                                            'Collaboration');
        if (sessionName) {
            let this_ = this;
            let errorCallback = function() {
                // FIXME
                alert("Failed to create session");
            }
            ko.collaboration.api.createSessionAsync(sessionName, function(sessionData) {
                if (sessionData && sessionData.hasOwnProperty("id")) {
                    this_.addSession(sessionData["id"], sessionData, true);
                } else {
                    errorCallback();
                }
            }, errorCallback);
        }
    },

    openSelectedDocument: function() {
        // FIXME disable on multi-select
        var index = this.treeView.getSelectedIndices()[0];
        if (typeof index === 'number') {
            let node = this.treeView._rows[index];
            if (node instanceof DocumentNode) {
                let filename = node.id;
                if (filename in mobwrite.shared) {
                    try {
                        mobwrite.shared[filename].view.makeCurrent();
                    } catch(e) {
                        log.error("Switching to " + filename + " failed: " +
                                  e.message);
                    }
                } else {
                    ko.collaboration.api.getTextAsync(filename, function(text) {
                        ko.collaboration.openExistingDocument(filename,
                            text.title, text.language, null);
                    }, function() {});
                }
            }
        }
    },

    addUserToSelectedSession: function(username) {
        // FIXME error handling
        var index = this.treeView.getSelectedIndices()[0];
        if (typeof index === 'number') {
            let sessionNode = this.treeView._rows[index];
            let successCallback =  ko.collaboration.utils.bind(this.refresh, this);
            if (username) {
                ko.collaboration.api.grantAccessAsync(username,
                                                   sessionNode.id,
                                                   successCallback);
            }
        }
    },

    addNewTabToSelectedSession: function() {
        var index = this.treeView.getSelectedIndices()[0];
        if (typeof index === 'number') {
            let sessionNode = this.treeView.getRootLevelNode(index);
            let title = ko.dialogs.prompt("Enter a name for this document", null, null, 'Collaboration');
            if (title) {
                let this_ = this;
                ko.collaboration.openNewDocument(sessionNode.id, title, null,
                    null, function() { this_.refresh(); } , function() {});
            }
        }
    },

    addExistingFileToSession: function() {
        var selection = this.treeView.getSelectedIndices()[0];
        var sessionNode = this.treeView.getRootLevelNode(selection);
        if (sessionNode instanceof SessionNode) {
            let this_ = this;
            let successCallback = function() {
                this_.refresh();
            };
            let errorCallback = function() {
                // TODO implement
            };
            let viewOpenedListener = function() {
                var view = ko.views.manager.currentView;
                ko.collaboration.openNewDocument(sessionNode.id, null, null,
                    view, successCallback, errorCallback);
            };
            ko.collaboration.utils.filePickerWithCallback(viewOpenedListener);
        }
    },

    deleteSelectedItems: function() {
        var selection = this.treeView.getSelectedIndices();
        for each (let row in selection) {
            // FIXME Should be able to pass in callback, that calls removeRow
            // (cp. removeSession) on success
            this.treeView.deleteRow(row);
        }
    },

    renameSelectedItem: function() {
        var selection = this.treeView.getSelectedIndices();
        // single selection only
        var selectedNode = this.treeView._rows[selection[0]];
        if (selectedNode instanceof SessionNode) {
            var title = ko.dialogs.prompt("Enter a new title for this session",
                null, selectedNode.name, 'Collaboration');
            if (title)
                ko.collaboration.api.updateSessionAsync(selectedNode.id, title,
                    successCallback, errorCallback);
        } else if (selectedNode instanceof DocumentNode) {
            var title = ko.dialogs.prompt("Enter a new title for this document",
                null, selectedNode.name, 'Collaboration');
            if (title) {
                ko.collaboration.api.updateTextAsync(selectedNode.id, title,
                    successCallback, errorCallback);
            }
        }
        function successCallback() {
            // Do nothing. Push notification should update the tree.
        }
        function errorCallback() {
            alert("Renaming failed");
        }
    },

    contextPopupHandler: function(evt) {
        // No guarantee that handling the click event is already finished and
        // selection has been corrected.
        if (evt.target == evt.currentTarget) {
            this.handleTreeClick(evt);
        }
        var selection = this.treeView.getSelectedIndices();
        // Note that we don't allow multiple selections. Empty selections are
        // possible though.
        var singleSelection = selection.length == 1;
        if (singleSelection) {
            var selectedNode = this.treeView._rows[selection[0]];
        }

        let elements = {
            "collab-sessions-context-open-document": {
                "hidden": !(singleSelection && selectedNode instanceof
                    DocumentNode)
            },
            "collab-sessions-context-add": {
                "disabled": !singleSelection    
            },
            "collab-sessions-context-create-session": {},
            "collab-sessions-context-rename": {
                "disabled": !singleSelection || !selectedNode.canBeRenamed()
            },
            "collab-sessions-context-delete": {
                "disabled" : !singleSelection || !selectedNode.canBeDeleted()
            }
        };
        for (let [id, att] in Iterator(elements)) {
            let element = document.getElementById(id);
            if (element) {
                for (let [k, v] in Iterator(att)) {
                    element.setAttribute(k, v);
                }
            }
        }
        return true;
    },
    
    addPopupHandler: function() {
        // Popup handler for the "Add..." popup in the context menu
        // Regular users can add documents, but adding users requires owner privileges.
        let element = document.getElementById("collab-sessions-context-add-user-menu");
        var selection = this.treeView.getSelectedIndices();
        var selectedNode = this.treeView._rows[selection[0]];
        let disabled = !selectedNode.hasOwnerPrivileges();
        element.setAttribute("disabled", disabled);
        return true;
    },

    userPopupHandler: function() {
        var menu = document.getElementById('collab-sessions-context-add-user-menu');
        var popup = document.getElementById('collab-sessions-context-add-user-popup');
        while (menu.itemCount > 0) {
            menu.removeItemAt(0);
        }
        if (!ko.collaboration.friends.allFriends ) {
            // List of friends unavailable
            let item = menu.appendItem("List of friends unavailable.");
            item.setAttribute('disabled', true);
        } else {
            // Display friends
            let noFriends = true;
            for each (let [userId, userName] in ko.collaboration.friends.getAllFriendsSorted()) {
                let item = menu.appendItem(userName, userId);
                let addUser = function(userId) {
                    return function() {
                        ko.collaboration.sessions.addUserToSelectedSession(userId);
                    }
                };
                item.addEventListener('click', addUser(userId), false);
                noFriends = false;
            }
            if (noFriends) {
                // Empty friends list
                let item = menu.appendItem("You need to add contacts to your friends list first.");
                item.setAttribute('disabled', true);
            }
        }
    },

    handleTreeClick: function(evt) {
        var row = this.tree.treeBoxObject.getRowAt(evt.pageX, evt.pageY);
        if (row >= 0)
            this.treeView.selection.select(row);
        else
            this.treeView.selection.clearSelection();
    },

    handleTreeDblClick: function(evt) {
        if (!evt || evt.which != 1) {
            // Disallow left-right double clicks on OS X (evt.which==3)
            return;
        }
        var rowIdx = this.tree.treeBoxObject.getRowAt(evt.pageX, evt.pageY);
        var row =  this.treeView._rows[rowIdx];
        if (row instanceof DocumentNode) {
            // This row should be selected, but just to be sure...
            this.treeView.selection.select(rowIdx);
            this.openSelectedDocument();
        }
    }
}

window.addEventListener("load", function() { ko.collaboration.sessions.init() }, false);
